/* SIP extension for UDP NAT alteration.
 *
 * (C) 2006 
 * based on RR's ip_nat_sip.c and other modules.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/ctype.h>
#include <linux/module.h>
#include <linux/netfilter_ipv4.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <net/udp.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_nat_helper.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_sip.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("sercomm suzhou <sercomm@sdc.sercomm.com>");
MODULE_DESCRIPTION("SIP NAT helper");

#define MAX_PORTS 8
static int ports[MAX_PORTS];
static int ports_c = 0;

#ifdef MODULE_PARM
MODULE_PARM(ports, "1-" __MODULE_STRING(MAX_PORTS) "i");
#endif


static DECLARE_LOCK(sip_alg_table_lock);

static unsigned int ip_ct_sip_timeout_bye = 2;
static unsigned int ip_ct_sip_timeout_cancel = 2;
static unsigned int ip_ct_sip_timeout_invite =64;

DECLARE_LOCK_EXTERN(sipbf_lock);

LIST_HEAD(sip_rtp_list);


#define G_CHANGE

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

#if 0
#define DEBUGP_G printk
#else
#define DEBUGP_G(format, args...)
#endif

/*extern void ip_nat_set_timeout_by_master(struct ip_conntrack * master,
								unsigned int timeout,
								struct ip_conntrack_tuple *tuple,
								struct ip_conntrack_tuple *mask);*/
extern void ip_ct_remove_expectations(struct ip_conntrack *ct);
extern struct sip_header_nfo ct_sip_hdrs[];
static unsigned int ip_nat_sdp(struct sk_buff **pskb, 
			enum ip_conntrack_info ctinfo,
			struct ip_conntrack_expect *exp,
			const char *dptr,
			struct ip_conntrack *ct,
			u_int16_t *ret_port);


static void expect_sdp_timed_out(unsigned long ul_expect)
{
	struct ip_contrack_sip_rtp *sip_expect = (void *) ul_expect;

	DEBUGP_G("!!!!!!!!!!!!!!!!  invite expectation %p timed out!!!!!!!!!!!!!!!!!!\n", sip_expect);	

	LOCK_BH(&sip_alg_table_lock);
	DEBUGP_G("unexpect sip_expect(%p)\n", sip_expect);
	/* delete from global and local lists */
	if(del_timer(&sip_expect->timeout))
		DEBUGP_G("because timeout , Success !!! before delete this sip_item . delete the timer for sip_del_item = %p\n", sip_expect);
	list_del(&sip_expect->list);
	kfree(sip_expect);
	UNLOCK_BH(&sip_alg_table_lock);
}



static unsigned int
sip_nat_expected(struct sk_buff **pskb,
		 unsigned int hooknum,
		 struct ip_conntrack *ct,
		 struct ip_nat_info *info)
{
	struct ip_nat_multi_range mr;
	u_int32_t newdstip, newsrcip, newip, mysrcip, mydstip;
	struct ip_ct_sip_expect *exp_sip_info;

	struct ip_conntrack *master = master_ct(ct);
	DEBUGP_G("************before  snat_expected: We have a connection!**************\n");
	
	IP_NF_ASSERT(info);
	IP_NF_ASSERT(master);

	IP_NF_ASSERT(!(info->initialized & (1<<HOOK2MANIP(hooknum))));

	DEBUGP_G("************nat_expected: We have a connection!**************\n");
	LOCK_BH(&sipbf_lock);

	exp_sip_info = &ct->master->help.exp_sip_info;

		newdstip = master->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
		newsrcip = master->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip;
		DEBUGP_G("nat_expected: RTP cmd. %u.%u.%u.%u->%u.%u.%u.%u\n",
		       NIPQUAD(newsrcip), NIPQUAD(newdstip));
		
	mysrcip = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
	mydstip = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip;

	if (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC)
		newip = newsrcip;
	else
		newip = newdstip;
	
	if((*pskb)->dev == NULL)
	printk("(*pskb)->dev = NULL \n");
	else if((*pskb)->dev->name == NULL)
	printk("(*pskb)->dev->name = NUll \n");
	else
	;

	if((*pskb)->dev != NULL)
	if (HOOK2MANIP(hooknum) == IP_NAT_MANIP_DST && memcmp((*pskb)->dev->name, "br0", 3) == 0 ) 
	{
		char devname[10];
		strncpy(devname, (*pskb)->dev->name, 9);
		devname[9] = 0;
		printk("devname = %s\n",devname);
		exp_sip_info->ipaddr_sip = mydstip;
	}



	if (HOOK2MANIP(hooknum) != IP_NAT_MANIP_DST) 
	{

		if(exp_sip_info->ipaddr_sip == mydstip)
		newip = mydstip;
		else
		newip = mysrcip;
		DEBUGP_G("nat sip change newip to mysrcip %u.%u.%u.%u\n", NIPQUAD(mysrcip));
	}
	DEBUGP_G("nat_expected: IP to %u.%u.%u.%u\n", NIPQUAD(newip));


	mr.rangesize = 1;
	/* We don't want to manip the per-protocol, just the IPs... */
	mr.range[0].flags = IP_NAT_RANGE_MAP_IPS;
	mr.range[0].min_ip = mr.range[0].max_ip = newip;

	/* ... unless we're doing a MANIP_DST, in which case, make
	   sure we map to the correct port */
	if (HOOK2MANIP(hooknum) == IP_NAT_MANIP_DST) {
		mr.range[0].flags |= IP_NAT_RANGE_PROTO_SPECIFIED;
		mr.range[0].min = mr.range[0].max
			= ((union ip_conntrack_manip_proto)
				{ htons(exp_sip_info->port) });
		DEBUGP_G("IP_NAT_MANIP_DST   dst port = %d\n",exp_sip_info->port);
	}
	UNLOCK_BH(&sipbf_lock);
	
	return ip_nat_setup_info(ct, &mr, hooknum);
}

/* So, this packet has hit the connection tracking matching code.
   and change the expectation to match the new version. just used for contact and via*/
static unsigned int ip_nat_sip_via_contact(struct sk_buff **pskb, 
			enum ip_conntrack_info ctinfo,
			struct ip_conntrack_expect *exp,
			const char *dptr,
			struct ip_conntrack *ct,
			u_int16_t *ret_port)
{
	u_int32_t newip;
	u_int16_t port;
	struct ip_conntrack_tuple newtuple;

	DEBUGP_G("ip_nat_sip_via_contact():\n");

	*ret_port = 0;
	
	/* Connection will come from reply */
	newip = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;
	
	memcpy(&newtuple, &exp->tuple, sizeof(exp->tuple));
	newtuple.dst.ip = newip;

	/* Try to get same port: if not, try to change it. */
	for (port = exp->help.exp_sip_info.port; port != 0; port++) 
	{
		newtuple.dst.u.udp.port = htons(port);

		{
			DEBUGP_G("sip gordon conntrack_sip: newtuple tuple: src=%u,%u,%u,%u:%u  dst= %u.%u.%u.%u:%u  proto=%d\n",
			       NIPQUAD(newtuple.src.ip),ntohs(newtuple.src.u.all),
			       NIPQUAD(newtuple.dst.ip),ntohs(newtuple.dst.u.all),newtuple.dst.protonum);
			DEBUGP_G("sip gordon conntrack_sip: exp->tuple: src=%u,%u,%u,%u:%u   dst= %u.%u.%u.%u:%u proto=%d\n",
			       NIPQUAD(exp->tuple.src.ip),ntohs(exp->tuple.src.u.all),
			       NIPQUAD(exp->tuple.dst.ip),ntohs(exp->tuple.dst.u.all),exp->tuple.dst.protonum);
		}
		if(ip_conntrack_change_expect(exp, &newtuple) == 0)
			break;
	}
	
	if (port == 0) 
	{
		return NF_DROP;
	}

	*ret_port = port;

	DEBUGP_G("ip_nat_sip_via_contact():  new_ret_port = %d\n",port);

	return NF_ACCEPT;
}

/*include rtp , contact , via*/
static int set_expected_packet(struct sk_buff **pskb,
			struct ip_conntrack *ct,
			enum ip_conntrack_info ctinfo, 
			u_int32_t ipaddr, u_int16_t port,
			const char *dptr,
			unsigned char flag_port,
			unsigned char current_flag,
			u_int16_t *ret_port,
			char *call_id)
{
	struct ip_conntrack_expect expect, *exp = &expect;
	struct ip_ct_sip_expect *exp_sip_info = &exp->help.exp_sip_info;
	int ret = NF_ACCEPT;
	int expect_ret = 0;
	
	
	memset(&expect, 0, sizeof(expect));

	exp->tuple = ((struct ip_conntrack_tuple)
		{ { ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip, { 0 } },
		  { ipaddr,  { htons(port) }, IPPROTO_UDP }});
		  
	exp->mask = ((struct ip_conntrack_tuple) 
		{ { 0, { 0 } },
		  { 0xFFFFFFFF, { .udp = { 0xFFFF } }, 0xFF }});

	exp->expectfn = NULL;
	exp_sip_info->port = port;
	exp_sip_info->ipaddr_sip=0;
	exp_sip_info->flag_port_contant = flag_port;
	exp_sip_info->current_flag = current_flag;
	exp_sip_info->contact_port = port;
	
	exp->master = ct;
	
	if(current_flag & CONTACT_FLAG)
		exp->reserved_flag = 1;

	if( current_flag & VIA_FLAG )
		exp->reserved_flag = 2;
#if 0
	if( current_flag & MEDIA_FLAG )/*if the except has arrived, but we haven't received ack , so we can't set the expectation again.*/
	{
		struct list_head *cur_item, *sip_item_temp = NULL;
		struct ip_contrack_sip_rtp *sip_item = NULL;
		struct ip_contrack_sip_rtp *sip_del_item = NULL;
		int call_id_length = 0;
		
		call_id_length = strlen(call_id);
		LOCK_BH(&sip_alg_table_lock);
		list_for_each_safe(cur_item, sip_item_temp, &sip_rtp_list) 
		{ 
			sip_item = list_entry(cur_item, struct ip_contrack_sip_rtp, list);
			DEBUGP_G("befor add set expect  this sip_item = %p\n", sip_item);
			if( call_id_length == sip_item->callid_length 
				&& strncmp(call_id, sip_item->call_id, call_id_length) == 0
				&& ip_ct_tuple_mask_cmp( &exp->tuple, &sip_item->tuple, &sip_item->mask))
			{
				DEBUGP_G("it is a same packet retransmitter before set expect !! Gordon this matched sip_item = %p\n", sip_item);
				if (!mangle_sdp(pskb, ctinfo, ct, newip, port, dptr)) 
				{
					return NF_DROP;
				}
				UNLOCK_BH(&sip_alg_table_lock);
				return NF_ACCEPT;
			}

		}
		UNLOCK_BH(&sip_alg_table_lock);
	}
#endif

	DEBUGP_G("before ip_conntrack_expect_related  exp = %p\n",exp);		
	expect_ret = ip_conntrack_expect_related(ct, exp);
	DEBUGP_G("after expect port = %d , expect_ret =  %d\n",port, expect_ret); 

	if(exp->ret_expect != NULL)
	{
		exp = exp->ret_expect;
		printk("ip_conntrack_expect_related success! exp = %p\n",exp);
	}
	else
	{
		printk(KERN_ERR"ERROR in ip_nat_sip.c function set_expected_packet() line:%d. it can add this expect in the expect list!\n",__LINE__);
		if(-EEXIST == expect_ret)
		{
			printk(KERN_ERR"Reason: This expectation has already exist!\n");
		}
		else if(-EPERM == expect_ret)
		{
			printk(KERN_ERR"Reason: This operation is not permitted\n");
		}
		else if(-EBUSY == expect_ret)
		{
			printk(KERN_ERR"Reason: This operation is busy now!\n");
		}
		else
		{
			printk(KERN_ERR"Reason: unknown reason expect_ret = %d\n",expect_ret);
		}
	}	
	

	DEBUGP_G("after ip_conntrack_expect_related  exp = %p\n",exp);		

	if(current_flag & CONTACT_FLAG || current_flag & VIA_FLAG)
	{
		ret = ip_nat_sip_via_contact(pskb, ctinfo, exp, dptr, ct, ret_port);
	}
	else if ( current_flag & MEDIA_FLAG ) 
	{
		ret = ip_nat_sdp(pskb, ctinfo, exp, dptr, ct, ret_port);
	}	


	return 	ret;
}

static int sip_help_bye(struct ip_conntrack *ct, char *call_id)
{
	int ret = NF_ACCEPT;
	struct list_head *cur_item,*sip_item_temp = NULL;
	struct ip_contrack_sip_rtp *sip_item = NULL;
	struct ip_contrack_sip_rtp *sip_del_item = NULL;
	int call_id_length = 0;
	
	ip_ct_remove_expectations(ct);
	
	call_id_length = strlen(call_id);
	
	DEBUGP_G("receive bye will remove the expectations and delete the RTP contract for this session\n");

		LOCK_BH(&sip_alg_table_lock);

		DEBUGP_G("sip_rtp_list = %p\n",&sip_rtp_list);
		list_for_each_safe(cur_item, sip_item_temp, &sip_rtp_list) 
		{
			DEBUGP_G("cur_item = %p\n",cur_item);
			sip_item = list_entry(cur_item, struct ip_contrack_sip_rtp, list);
			DEBUGP_G("this sip_item = %p, sip_item->list.prev = %p, sip_item->list.next = %p\n", sip_item, sip_item->list.prev, sip_item->list.next);
			
		}	
		
		DEBUGP_G("sip_rtp_list = %p  out of first cycle\n",&sip_rtp_list);
		
		list_for_each_safe(cur_item, sip_item_temp, &sip_rtp_list) 
		{ 
			DEBUGP_G("cur_item = %p\n",cur_item);
			sip_item = list_entry(cur_item, struct ip_contrack_sip_rtp, list);
			DEBUGP_G("this sip_item = %p\n", sip_item);
			DEBUGP_G("readed  sip_item = %p, sip_item->list.prev = %p, sip_item->list.next = %p\n", sip_item, sip_item->list.prev, sip_item->list.next);
			if(call_id_length == sip_item->callid_length && strncmp(call_id, sip_item->call_id, call_id_length) == 0)
			{
				sip_del_item = sip_item;
				DEBUGP_G("Gordon this matched sip_item = %p\n", sip_item);
				ip_nat_set_timeout_by_master(ct, ip_ct_sip_timeout_bye, &sip_item->tuple, &sip_item->mask);
			}
			if(sip_del_item)
			{
				DEBUGP_G("delete this list Gordon this matched sip_del_item = %p\n", sip_del_item);
				if(del_timer(&sip_del_item->timeout))
					DEBUGP_G("Success !!! before delete this sip_item . delete the timer for sip_del_item = %p\n", sip_del_item);
				
				DEBUGP_G("before delete sip_del_item = %p, sip_del_item->list.prev = %p, sip_del_item->list.next = %p\n", sip_del_item, sip_del_item->list.prev, sip_del_item->list.next);
				
				list_del_init(&sip_del_item->list);
				kfree(sip_del_item);
				
				DEBUGP_G("after delete sip_del_item = %p, sip_del_item->list.prev = %p, sip_del_item->list.next = %p\n", sip_del_item, sip_del_item->list.prev, sip_del_item->list.next);
				
				sip_del_item = NULL;
			}
		
			DEBUGP_G("delete this item  sip_del_item = %p success!\n", sip_del_item);
			
		}
		UNLOCK_BH(&sip_alg_table_lock);
	
	return ret;
}

static int sip_help_cancel(struct ip_conntrack *ct)
{
	int ret = NF_ACCEPT;
	ip_ct_remove_expectations(ct);
	ip_nat_set_timeout_by_master(ct, ip_ct_sip_timeout_cancel, NULL, NULL);
	return ret;
}

static int sip_help_bad(struct ip_conntrack *ct)
{
	int ret = NF_ACCEPT;
	ip_ct_remove_expectations(ct);
	return ret;
}

int 
is_200_for_require(const char *s_dptr, unsigned int datalen, char *str_target)
{
	int ret = 0;
	if(memcmp(s_dptr, "SIP/2.0 200", sizeof("SIP/2.0 200") - 1) == 0 || (memcmp(s_dptr, "ACK", sizeof("ACK") - 1) == 0))	
	{
			const char *for_200 = ct_sip_search("CSeq:", s_dptr, sizeof("CSeq:") - 1, 
									datalen);
			if (!for_200)
			{
				DEBUGP_G("search CSeq fail ,test Cseq\n");
				for_200 = ct_sip_search("Cseq:", s_dptr, sizeof("Cseq:") - 1, 
									datalen);
				if (!for_200)
				{
					DEBUGP_G("search Cseq fail return 0\n");
					ret = 0;
				}	
			}
		
			if (ct_sip_search(str_target, for_200, strlen(str_target), 
			    ct_sip_lnlen(for_200, s_dptr + datalen))) 
			{
					DEBUGP_G("search %s success return 1\n", str_target);
					ret = 1;
			}
	}
	return ret;
}

#define G_ADD
/* Simple ipaddr parser.. */
static int parse_ipaddr(const char *cp,	const char **endp, 
			u_int32_t *ipaddr, const char *limit)
{
	unsigned long int val;
	int i, digit = 0;
	
	for (i = 0, *ipaddr = 0; cp <= limit && i < 4; i++) 
	{
		digit = 0;
		if (!isdigit(*cp))
			break;
		
		val = simple_strtoul(cp, (char **)&cp, 10);
		if (val > 0xFF)
			return -1;
	
		((uint8_t *)ipaddr)[i] = val;	
		digit = 1;
	
		if (*cp != '.')
			break;
		cp++;
	}
	if (!digit)
		return -1;
	
	if (endp)
		*endp = cp;

	return 0;
}



static unsigned int mangle_sip_packet(struct sk_buff **pskb, 
			enum ip_conntrack_info ctinfo,
			struct ip_conntrack *ct, 
			const char **dptr, size_t dlen,
			char *buffer, int bufflen,
			struct sip_header_nfo *hnfo)
{
	unsigned int matchlen, matchoff;
	u_int16_t port_useless;
#ifdef G_ADD
	const char *s_dptr;
	int dir;
	char buffer_s[sizeof("nnn.nnn.nnn.nnn:nnnnn")];	
	unsigned int bufflen_s;
	int i=0;
	u_int32_t ip;
	u_int16_t port;
#endif	

	if (ct_sip_get_info(*dptr, dlen, &matchoff, &matchlen, hnfo, &port_useless) <= 0)
		return 0;

#ifdef G_ADD		
	/*if the original src ip = newip, we need to change the ip, else do nothing*/	
	s_dptr = *dptr;
	dir = CTINFO2DIR(ctinfo);
	if ((ctinfo) >= IP_CT_IS_REPLY) 
	{
		ip = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;
		port = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u.udp.port;
	} else 
	{
		ip = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
		port = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port;
	}
	bufflen_s = sprintf(buffer_s, "%u.%u.%u.%u", NIPQUAD(ip));

			DEBUGP("source packet data:");
			for(i=0; i<3; i++)
			DEBUGP("%c",*(s_dptr + matchoff+i));
			DEBUGP("\nneed change to:");
			for(i=0; i<3; i++)
			DEBUGP("%c",*(buffer_s+i));
			DEBUGP("\ncompare ip , whether need change packet?\n");
	if(memcmp(hnfo->lname, "\nm=", sizeof("\nm=")) != 0 
		&& memcmp(hnfo->ln_str, "req_src_port=", sizeof("req_src_port=")) != 0 
		&& memcmp(s_dptr + matchoff, buffer_s, bufflen_s) != 0)
		{
			
			DEBUGP("source packet data:");
			for(i=0; i<3; i++)
			DEBUGP("%c",*(s_dptr + matchoff+i));
			DEBUGP("\nneed change to:");
			for(i=0; i<3; i++)
			DEBUGP("%c",*(buffer+i));
			DEBUGP("\nno the private ip , so needn't change packet\n");
			goto out;	
		}
#endif

	if (!ip_nat_mangle_udp_packet(pskb, ct, ctinfo, 
	    matchoff, matchlen, buffer, bufflen)) 
	{
		return 0;
	}
	/* We need to reload this. Thanks Patrick. */
	out:
	*dptr = (*pskb)->data + (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr); 
	return 1;
}


static unsigned int ip_nat_sip(struct ip_conntrack *ct,
			 struct ip_conntrack_expect *exp,
			 struct ip_nat_info *info,
			 enum ip_conntrack_info ctinfo,
			 unsigned int hooknum,
			 struct sk_buff **pskb)
{
	char buffer[sizeof("nnn.nnn.nnn.nnn:nnnnn")];	
	char buffer_contact[sizeof("nnn.nnn.nnn.nnn:nnnnn")];	
	char buffer_via[sizeof("nnn.nnn.nnn.nnn:nnnnn")];	
	char buffer_warning_src_port[sizeof("nnnnn")];	
	char buffer_warning_src_ip[sizeof("nnn.nnn.nnn.nnn")];	
	unsigned int bufflen, bufflen_contact, bufflen_via, dataoff,buffer_warning_src_port_len,buffer_warning_src_ip_len;
	u_int32_t ip;
	u_int16_t port, ret_port, via_port;
	int dir;
	const char *dptr,*s_dptr;
	int ret = NF_ACCEPT;
	struct ip_ct_sip_expect *exp_sip_info = NULL;

#ifdef G_CHANGE
	unsigned int udplen = (*pskb)->len - (*pskb)->nh.iph->ihl * 4;
	unsigned int datalen = udplen - sizeof(struct udphdr);
	int matchoff, matchlen;
	u_int32_t src_ip;
	char str_invite[] = "INVITE";
	char str_bye[] = "BYE";
	char str_ack[] = "ACK";

	/*we need got Call-ID*/
	char call_id[128] = {'\0'};
	u_int16_t ret_connect_port = 0;
	int call_id_length = 0 ;
	
	u_int32_t ipaddr,ipaddr_contact,ipaddr_via;
	u_int16_t sdp_port=0,ret_sdp_port=0,ret_contact_port=0,ret_via_port=0,src_port=0;
#endif	

	DEBUGP_G("gordon enter ip_nat_sip\n");	
	if(exp)
		exp_sip_info = &exp->help.exp_sip_info;
	dataoff = (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr);
	if (dataoff >= (*pskb)->len) {
		DEBUGP("skb->len = %u\n", (*pskb)->len);
		return NF_ACCEPT;
        }
	dptr = (*pskb)->data + dataoff;
	s_dptr = dptr;


	dir = CTINFO2DIR(ctinfo);
	if (!((hooknum == NF_IP_POST_ROUTING && dir == IP_CT_DIR_ORIGINAL)
	      || (hooknum == NF_IP_PRE_ROUTING && dir == IP_CT_DIR_REPLY))) {
		DEBUGP_G("nat_sip: Not touching dir %s at hook %s\n",
		       dir == IP_CT_DIR_ORIGINAL ? "ORIG" : "REPLY",
		       hooknum == NF_IP_POST_ROUTING ? "POSTROUTING"
		       : hooknum == NF_IP_PRE_ROUTING ? "PREROUTING"
		       : hooknum == NF_IP_LOCAL_OUT ? "OUTPUT" : "???");
		return NF_ACCEPT;
	}
	
	if ((ctinfo) >= IP_CT_IS_REPLY) {
		ip = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
		port = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port;
		src_ip = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;
		src_port = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u.udp.port;
	} else {
		ip = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;
		port = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u.udp.port;
		src_ip = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
		src_port = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port;
	}
	
	/*Gordon add for change the port in the contact and via field*/
	
//	via_port = exp->help.exp_sip_info.via_port;
	bufflen = sprintf(buffer, "%u.%u.%u.%u:%u", NIPQUAD(ip), ntohs(port));
	DEBUGP("sip buffer=%s\n",buffer);	
	buffer_warning_src_port_len = sprintf(buffer_warning_src_port, "%u", ntohs(port));
	DEBUGP("buffer_warning_src_port=%s\n",buffer_warning_src_port);	
	buffer_warning_src_ip_len = sprintf(buffer_warning_src_ip, "%u.%u.%u.%u", NIPQUAD(ip));
	DEBUGP("buffer_warning_src_ip=%s\n",buffer_warning_src_ip);	
	
	
	/* short packet ? */
	if (((*pskb)->len - dataoff) < (sizeof("SIP/2.0") - 1))
		return NF_DROP;
	
		LOCK_BH(&sipbf_lock);

#ifdef G_CHANGE
		/*we distingust the packet by the direction*/	
		/*the following if else just generate the expection if needed.*/


	
	{
			/*now we will set flag for contact and via header*/
		unsigned char flag_port = 0;

		if (ct_sip_get_info(dptr, datalen, &matchoff, &matchlen, 
		    &ct_sip_hdrs[POS_CALL_ID_HEADER], &ret_connect_port) > 0)
		{
			if(matchlen >127)
				matchlen = 127;
				
			strncpy( call_id, dptr + matchoff, matchlen);
			call_id_length = matchlen;
			DEBUGP_G("this packet's call-id: %s\n",call_id);
		}
		
		
		if( (ctinfo) < IP_CT_IS_REPLY  )
		{
			/*we need to do 3 things. 
			1.check the contact header, and decide if we need setup the expection.
			2.check the via header, and decide if we need setup the expection.
			3.check whether we need change the sdp content.*/
			
			DEBUGP_G("it it a outgoing packet, maybe with sdp ,diff via and diff contact\n");
			/*now first thing.*/
			if (ct_sip_get_info(dptr, datalen, &matchoff, &matchlen, 
			    &ct_sip_hdrs[POS_CONTACT], &ret_contact_port) > 0) 
			{
				ipaddr_contact = 0;
				if (parse_ipaddr(dptr + matchoff, NULL, &ipaddr_contact, 
				    dptr + datalen) < 0) 
				{
					ret = NF_DROP;
					goto exit;
				}
				if (0 != ret_contact_port )
				{
					if (ret_contact_port < 1024) 
					{
						ret = NF_DROP;
						goto exit;
					}
					
					if(src_ip == ipaddr_contact)
					{
						flag_port |= CONTACT_FLAG;	
						DEBUGP_G("this packet's contact field  include port %d,contact ip= %u.%u.%u.%u,src_port = %d\n",ret_contact_port,NIPQUAD(ipaddr_contact),ntohs(src_port)); 
					}
			
					if((ret_contact_port != ntohs(src_port)) && (src_ip == ipaddr_contact))
					{
						DEBUGP_G("outgoing packet will do expect contact \n");
						ret = set_expected_packet(pskb, ct, ctinfo,
								ipaddr_contact, ret_contact_port, dptr, flag_port, CONTACT_FLAG, &ret_contact_port, call_id);
						if( ret != NF_ACCEPT)
							goto  exit;
						flag_port |= CONTACT_IN_EXP;
					}
				}
		
			}
			
			
			/*second thing*/
			/* Get ip and port address from SIP Via header*/
			if (ct_sip_get_info(dptr, datalen, &matchoff, &matchlen, 
			    &ct_sip_hdrs[POS_VIA], &ret_via_port) > 0) 
			{
				ipaddr_via = 0;
		
				if (parse_ipaddr(dptr + matchoff, NULL, &ipaddr_via, 
				    dptr + datalen) < 0) 
				{
					ret = NF_DROP;
					goto exit;
				}
					DEBUGP("via field  src_ip=%u.%u.%u.%u , via contant ip = %u.%u.%u.%u\n",NIPQUAD(ip),NIPQUAD(ipaddr_via)); 
						
				if (0 != ret_via_port )
				{
					if (ret_via_port < 1024) 
					{
						ret = NF_DROP;
						goto exit;
					}
					
					if(src_ip == ipaddr_via)
					{
						flag_port |= VIA_FLAG;
						DEBUGP_G("this packet's via field  include port %d, src_port=%d\n",ret_via_port,ntohs(src_port)); 
					}

					if( (ret_via_port != ntohs(src_port)) && (src_ip == ipaddr_via) )
					{
						DEBUGP_G("outgoing packet will do expect via \n");
						ret = set_expected_packet(pskb, ct, ctinfo,
								ipaddr_via, ret_via_port, dptr, flag_port, VIA_FLAG, &ret_via_port, call_id);

						if( ret != NF_ACCEPT)
							goto  exit;
				
						flag_port |= VIA_IN_EXP;
					}
				}

			}
		
			
			if( ret != NF_ACCEPT)
				goto  exit;

			DEBUGP_G("this packet's contact field  port=%d,udp packet src port=%d, via field port=%d\n",ret_contact_port,ntohs(src_port),ret_via_port); 
			
			/*third thing , but we must confirm this is a outgoing INVITE or 200 OK in the first place */
			
				/* RTP info only in some SDP pkts */
			if (memcmp(dptr, "INVITE", sizeof("INVITE") - 1) == 0 || 
			    (memcmp(dptr, "SIP/2.0 200", sizeof("SIP/2.0 200") - 1) == 0 && is_200_for_require(dptr, datalen, str_invite) ) )

				/* Get ip and port address from SDP packet. */
			if (ct_sip_get_info(dptr, datalen, &matchoff, &matchlen, 
			    &ct_sip_hdrs[POS_CONECTION], &ret_sdp_port) > 0) 
			{
		
				/* We'll drop only if there are parse problems. */
				if (parse_ipaddr(dptr + matchoff, NULL, &ipaddr, 
				    dptr + datalen) < 0) 
				{
					ret = NF_DROP;
					goto exit;
				}
				
				/*think of re-invite the ip in the sdp is 0, we delete the expection*/
				if(ipaddr == 0 )
				{
					DEBUGP_G("send the 200 OK with 0 ip in sdp, we will delete the expection\n");
					sip_help_bye(ct, call_id);
				}
				else
				{
					if (ct_sip_get_info(dptr, datalen, &matchoff, &matchlen, 
					    &ct_sip_hdrs[POS_MEDIA], &ret_sdp_port) > 0) 
					{
			
						sdp_port = simple_strtoul(dptr + matchoff, NULL, 10);
						if (sdp_port < 1024) 
						{
							ret = NF_DROP;
							goto exit;
						}
							DEBUGP_G("outgoing packet will do expect sdp \n");
			
						ret = set_expected_packet(pskb, ct, ctinfo,
									ipaddr, sdp_port, dptr, flag_port, MEDIA_FLAG, &ret_sdp_port, call_id);
						if( ret != NF_ACCEPT)
							goto  exit;
					}
				}
			}
				
					
		}
		else /*(ctinfo) >= IP_CT_IS_REPLY */
		{
			/*need do nothing for expectation.  But we need set the flag for contact and via,
			 just set flag without expectation*/
			
			
			/* two things need do
			1. check contact header.
			2. check via header.
			*/
			DEBUGP_G("it is a incoming packet , just set the flag for contact and via\n");
			/*now first thing.*/
			if (ct_sip_get_info(dptr, datalen, &matchoff, &matchlen, 
			    &ct_sip_hdrs[POS_CONTACT], &ret_contact_port) > 0) 
			{
				ipaddr_contact = 0;
				if (parse_ipaddr(dptr + matchoff, NULL, &ipaddr_contact, 
				    dptr + datalen) < 0) 
				{
					ret = NF_DROP;
					goto exit;
				}
				if (0 != ret_contact_port )
				{
					if (ret_contact_port < 1024) 
					{
						ret = NF_DROP;
						goto exit;
					}
					
					if(src_ip == ipaddr_contact)
					{
						flag_port |= CONTACT_FLAG;	
						DEBUGP_G("this packet's contact field  include port %d,contact ip= %u.%u.%u.%u\n",ret_contact_port,NIPQUAD(ipaddr_contact)); 
					}
			
				}
		
			}
			
			
			/*second thing*/
			
			/* Get ip and port address from SIP Via header*/
			if (ct_sip_get_info(dptr, datalen, &matchoff, &matchlen, 
			    &ct_sip_hdrs[POS_VIA], &ret_via_port) > 0) 
			{
				ipaddr_via = 0;
		
				if (parse_ipaddr(dptr + matchoff, NULL, &ipaddr_via, 
				    dptr + datalen) < 0) 
				{
					ret = NF_DROP;
					goto exit;
				}
					DEBUGP("via field  src_ip=%u.%u.%u.%u , via contant ip = %u.%u.%u.%u\n",NIPQUAD(ip),NIPQUAD(ipaddr_via)); 
						
				if (0 != ret_via_port )
				{
					if (ret_via_port < 1024) 
					{
						ret = NF_DROP;
						goto exit;
					}
					
					if(src_ip == ipaddr_via)
					{
						flag_port |= VIA_FLAG;	
						DEBUGP_G("this packet's via field  include port %d\n",ret_via_port); 
					}
				}

			}
			
			/*if it's incoming 200 OK for INVITE or ACK for INVITE's 200 OK, we need del the timer */
			if ((memcmp(dptr, "ACK", sizeof("ACK") - 1) == 0 && is_200_for_require(dptr, datalen, str_ack))
				|| (memcmp(dptr, "SIP/2.0 200", sizeof("SIP/2.0 200") - 1) == 0 && is_200_for_require(dptr, datalen, str_invite) ))
			{
				struct list_head *cur_item, *sip_item_temp = NULL;
				struct ip_contrack_sip_rtp *sip_item = NULL;
				struct ip_contrack_sip_rtp *sip_del_item = NULL;
				
				LOCK_BH(&sip_alg_table_lock);
				DEBUGP_G("it is a ACK for INVITE's 200 OK or incoming 200 OK for outgoing INVITE \n");
				DEBUGP_G("sip_rtp_list = %p\n",&sip_rtp_list);
				list_for_each_safe(cur_item, sip_item_temp, &sip_rtp_list) 
				{
					DEBUGP_G("cur_item = %p\n",cur_item);
					
				}	
					DEBUGP_G("sip_rtp_list = %p  out of first cycle\n",&sip_rtp_list);
				
				list_for_each_safe(cur_item, sip_item_temp, &sip_rtp_list) 
				{ 
					DEBUGP_G("cur_item = %p\n",cur_item);
					sip_item = list_entry(cur_item, struct ip_contrack_sip_rtp, list);
					DEBUGP_G("this sip_item = %p\n", sip_item);
					if(call_id_length == sip_item->callid_length && strncmp(call_id, sip_item->call_id, call_id_length) == 0)
					{
						sip_del_item = sip_item;
						DEBUGP_G("Gordon this matched sip_item = %p, we will delete the timer\n", sip_item);
					}
					if(sip_del_item)
					{
						DEBUGP_G("delete the timer Gordon this matched sip_del_item = %p\n", sip_del_item);
						if(del_timer(&sip_del_item->timeout))
						{
							DEBUGP_G("Success !!! delete the timer for sip_del_item = %p\n", sip_del_item);
						}
						sip_del_item = NULL;
					}
						DEBUGP_G("delete the timer for  sip_del_item = %p  success\n", sip_del_item);
					
				}
				UNLOCK_BH(&sip_alg_table_lock);
				
			}
			
			
		
			DEBUGP_G("this packet's contact field  port=%d,udp packet src port=%d, via field port=%d\n",ret_contact_port,ntohs(src_port),ret_via_port); 
			
		}
		
		/*now we will change the contact, via , warning , and request header.*/
		{
			
			/*we have to do two thing , construct the contact header and via header */

			/*1. construct contact header.*/
			if(flag_port & CONTACT_FLAG)
			{
				if( flag_port & CONTACT_IN_EXP )
				{
					bufflen_contact = sprintf(buffer_contact, "%u.%u.%u.%u:%u", NIPQUAD(ip), ret_contact_port);
					DEBUGP("buffer_contact use the new contact port\n");
				}
				else /*use the udp port*/
				{
					bufflen_contact = sprintf(buffer_contact, "%u.%u.%u.%u:%u", NIPQUAD(ip), ntohs(port));
					DEBUGP("buffer_contact use the upd port\n");
				}
			}
			else /*without contact port*/
			{
				bufflen_contact = sprintf(buffer_contact, "%u.%u.%u.%u", NIPQUAD(ip));
				DEBUGP("buffer_contact without port\n");
			}
			DEBUGP_G("bufflen_contact = %s\n",buffer_contact);
			
			/*2. construct via header.*/
			if(flag_port & VIA_FLAG)
			{
				if( flag_port & VIA_IN_EXP )
				{
					bufflen_via = sprintf(buffer_via, "%u.%u.%u.%u:%u", NIPQUAD(ip), ret_via_port);
					DEBUGP("buffer_via use the new via port\n");
				}
				else /*use the udp port*/
				{
					bufflen_via = sprintf(buffer_via, "%u.%u.%u.%u:%u", NIPQUAD(ip), ntohs(port));
					DEBUGP("buffer_via use the upd port\n");
				}
			}
			else /*without via port*/
			{
				bufflen_via = sprintf(buffer_via, "%u.%u.%u.%u", NIPQUAD(ip));
				DEBUGP("buffer_via without port\n");
			}
			DEBUGP_G("buffer_via=%s\n",buffer_via);	
		}
	}	
		
		/*fill the correct via and contact header*/
		{
			/* Basic rules: requests and responses. */
			
			/*anything 401 100 180 200 must change. outgoing packet only need change contact field,
				incoming packet need change via field. if incoming packet include warning message 
				also need change. if the incoming packet 401 200 for REGISTER, we need change the 
				contact header too.*/
			if (memcmp(dptr, "SIP/2.0", sizeof("SIP/2.0") - 1) == 0) 
			{
				DEBUGP("first  dptr=%p\n",dptr);
				if ((ctinfo) < IP_CT_IS_REPLY) {
					mangle_sip_packet(pskb, ctinfo, ct, &dptr, 
						(*pskb)->len - dataoff, buffer_contact, bufflen_contact,
						&ct_sip_hdrs[POS_CONTACT]);
						goto next;
				}
		
				if (!mangle_sip_packet(pskb, ctinfo, ct, &dptr, (*pskb)->len - dataoff,
				    buffer_via, bufflen_via, &ct_sip_hdrs[POS_VIA]))
				{
					DEBUGP("mangle VIA goto exit\n");
					goto exit;
				}
			DEBUGP("second  dptr=%p\n",dptr);
				
				/*Gordon add to modify the response for REGISTER , for instance , 401 ,200 OK, also INVITE 100 try req_src_ip and req_src_port*/				
				mangle_sip_packet(pskb, ctinfo, ct, &dptr, (*pskb)->len - dataoff, 
								buffer_warning_src_ip, buffer_warning_src_ip_len, &ct_sip_hdrs[POS_WARNING_SRC_IP]);
								
				mangle_sip_packet(pskb, ctinfo, ct, &dptr, (*pskb)->len - dataoff, 
								buffer_warning_src_port, buffer_warning_src_port_len, &ct_sip_hdrs[POS_WARNING_SRC_PORT]);
		
				{
				/* This search should ignore case, but later.. */
				const char *aux = ct_sip_search("CSeq:", dptr, sizeof("CSeq:") - 1, 
										(*pskb)->len - dataoff);
				if (!aux)
				{
					DEBUGP("search CSeq fail ,test Cseq\n");
					aux = ct_sip_search("Cseq:", dptr, sizeof("Cseq:") - 1, 
										(*pskb)->len - dataoff);
					if (!aux)
					{
						DEBUGP("search Cseq fail, and CSeq fail before goto exit\n");
						goto exit;
					}
				}
			
				if (!ct_sip_search("REGISTER", aux, sizeof("REGISTER") - 1, 
				    ct_sip_lnlen(aux, dptr + (*pskb)->len - dataoff))) 
				{
					DEBUGP("search REGISTER fail goto next\n");
					goto next;
				}
				ret = mangle_sip_packet(pskb, ctinfo, ct, &dptr, (*pskb)->len - dataoff, 
								buffer_contact, bufflen_contact, &ct_sip_hdrs[POS_CONTACT]);
								
					goto next;
				}
					
			}
		
			/*anything outgoing must change via , if we have contact field ,need change also .except above situation, for example bye invite register cancel*/
			if ((ctinfo) < IP_CT_IS_REPLY) 
			{
				if (!mangle_sip_packet(pskb, ctinfo, ct, &dptr, (*pskb)->len - dataoff,
				    buffer_via, bufflen_via, &ct_sip_hdrs[POS_VIA])) 
				{
					goto exit;
				}
				
				/* Mangle Contact if exists only. - watch udp_nat_mangle()! */
				mangle_sip_packet(pskb, ctinfo, ct, &dptr, (*pskb)->len - dataoff, 
								buffer_contact, bufflen_contact, &ct_sip_hdrs[POS_CONTACT]);
				goto next;
			}
			
			/* This mangle requests headers. only for incoming invite register  and bye*/
				ret = mangle_sip_packet(pskb, ctinfo, ct, &dptr, 
					ct_sip_lnlen(dptr, dptr + (*pskb)->len - dataoff),
					buffer, bufflen, &ct_sip_hdrs[POS_REQ_HEADER]);
				if(!ret)
				{
					DEBUGP_G("mangle POS_REQ_HEADER failed go to exit\n");
					goto exit;
				}
				else
					goto next;
		}

	next:
		DEBUGP_G("come to next we will do the delete contrack and expection if needed\n");
		
		/*then we will do the last thing. for instance, delete the contrack and the expectation*/
		{
			if((ctinfo) < IP_CT_IS_REPLY && memcmp(s_dptr, "CANCEL ", sizeof("CANCEL ") - 1) == 0)	
			{
				DEBUGP_G("receive outgoing cancel\n");
				sip_help_cancel(ct);
				UNLOCK_BH(&sipbf_lock);
			  	return NF_ACCEPT;  
			}
			else if((ctinfo) >= IP_CT_IS_REPLY && memcmp(s_dptr, "SIP/2.0 400", sizeof("SIP/2.0 400") - 1) == 0)	
			{
				DEBUGP_G("receive incomig 400\n");
				sip_help_bad(ct);
				UNLOCK_BH(&sipbf_lock);
			  	return NF_ACCEPT;
			}	
//			else if(memcmp(s_dptr, "BYE ", sizeof("BYE ") - 2) == 0)
			else if((memcmp(s_dptr, "SIP/2.0 200", sizeof("SIP/2.0 200") - 2) == 0) && is_200_for_require(s_dptr, datalen, str_bye))
			{
				DEBUGP_G("receive BYE's 200 OK\n");
				sip_help_bye(ct, call_id);
				UNLOCK_BH(&sipbf_lock);
			  	return NF_ACCEPT;  
				
			}
			
			{
				UNLOCK_BH(&sipbf_lock);
				return NF_ACCEPT;
			}
		}
		
		/*the end*/
	exit:
		DEBUGP_G("come to exit\n");
		sip_help_bye(ct, call_id);
		UNLOCK_BH(&sipbf_lock);
		return NF_DROP;
#endif

}

static int mangle_content_len(struct sk_buff **pskb, 
			enum ip_conntrack_info ctinfo,
			struct ip_conntrack *ct,
			const char *dptr)
{
	unsigned int dataoff, matchoff, matchlen;
	char buffer[sizeof("65536")];
	int bufflen;
	u_int16_t port_useless;
	
	dataoff = (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr);
	
	/* Get actual SDP lenght */
	if (ct_sip_get_info(dptr, (*pskb)->len - dataoff, &matchoff, 
	    &matchlen, &ct_sip_hdrs[POS_SDP_HEADER], &port_useless) > 0) 
	{
		
		/* since ct_sip_get_info() give us a pointer passing 'v='
		   we need to add 2 bytes in this count. */
		int c_len = (*pskb)->len - dataoff - matchoff + 2;
		
		/* Now, update SDP lenght */
		if (ct_sip_get_info(dptr, (*pskb)->len - dataoff, &matchoff, 
		    &matchlen, &ct_sip_hdrs[POS_CONTENT], &port_useless) > 0) 
		{
		    
			bufflen = sprintf(buffer, "%u", c_len);
			
			return ip_nat_mangle_udp_packet(pskb, ct, ctinfo, matchoff,
							matchlen, buffer, bufflen);
		}
	}
	return 0;
}

static unsigned int mangle_sdp(struct sk_buff **pskb, 
			enum ip_conntrack_info ctinfo,
			struct ip_conntrack *ct,
			u_int32_t newip, u_int16_t port,
			const char *dptr)
{
	char buffer[sizeof("nnn.nnn.nnn.nnn")];
	unsigned int dataoff, bufflen;

	dataoff = (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr);
	
	/* Mangle owner and contact info. */
	bufflen = sprintf(buffer, "%u.%u.%u.%u", NIPQUAD(newip));
	DEBUGP_G("mangle_sdp sdp buffer=%s\n",buffer);
	if (!mangle_sip_packet(pskb, ctinfo, ct, &dptr, (*pskb)->len - dataoff,
	    buffer, bufflen, &ct_sip_hdrs[POS_OWNER])) 
	{
		return 0;	
	}
	
	if (!mangle_sip_packet(pskb, ctinfo, ct, &dptr, (*pskb)->len - dataoff,
	    buffer, bufflen, &ct_sip_hdrs[POS_CONECTION]))
	{
		return 0;
	}
	
	/* Mangle media port. */
	bufflen = sprintf(buffer, "%u", port);
	if (!mangle_sip_packet(pskb, ctinfo, ct, &dptr, (*pskb)->len - dataoff,
	    buffer, bufflen, &ct_sip_hdrs[POS_MEDIA])) 
	{
		return 0;
	}
	
	return mangle_content_len(pskb, ctinfo, ct, dptr);
}


/* So, this packet has hit the connection tracking matching code.
   Mangle it, and change the expectation to match the new version. */
static unsigned int ip_nat_sdp(struct sk_buff **pskb, 
			enum ip_conntrack_info ctinfo,
			struct ip_conntrack_expect *exp,
			const char *dptr,
			struct ip_conntrack *ct,
			u_int16_t *ret_port)
{
	u_int32_t newip;
	u_int16_t port;
	struct ip_conntrack_tuple newtuple;
	int matchoff, matchlen;
	u_int32_t ipaddr=0;
	u_int16_t ret_connect_port=0;
	unsigned int dataoff;
	struct list_head *cur_item, *sip_item_temp=NULL;
	struct ip_contrack_sip_rtp *sip_item = NULL;
	int change_ret = 0;
	DEBUGP_G("ip_nat_sdp():\n");


#if 0
	dataoff = (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr);

	if(ctinfo >= IP_CT_IS_REPLY)
			/* Get ip and port address from SDP packet. */
		if (ct_sip_get_info(dptr, (*pskb)->len - dataoff, &matchoff, &matchlen, 
		    &ct_sip_hdrs[POS_CONECTION], &ret_connect_port) > 0) 
		{
	
			/* We'll drop only if there are parse problems. */
			if (parse_ipaddr(dptr + matchoff, NULL, &ipaddr, 
			    dptr + (*pskb)->len - dataoff) < 0) 
			{
				return NF_DROP;
			}
			DEBUGP_G("ip_nat_sdp got 200 OK of invite reply ipaddr = %u,%u,%u,%u\n",NIPQUAD(ipaddr));
		}
#endif

	/* Connection will come from reply */
	newip = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;
	
	memcpy(&newtuple, &exp->tuple, sizeof(exp->tuple));
	newtuple.dst.ip = newip;

	exp->help.exp_sip_info.ipaddr_sip = 0;

	/* Try to get same port: if not, try to change it. */
	for (port = exp->help.exp_sip_info.port; port != 0; port++) 
	{
		
		newtuple.dst.u.udp.port = htons(port);

		{
			DEBUGP_G("gordon conntrack_sip: newtuple tuple: src=%u,%u,%u,%u:%u  dst= %u.%u.%u.%u:%u  proto=%d\n",
			       NIPQUAD(newtuple.src.ip),ntohs(newtuple.src.u.all),
			       NIPQUAD(newtuple.dst.ip),ntohs(newtuple.dst.u.all),newtuple.dst.protonum);
			DEBUGP_G("gordon conntrack_sip: exp->tuple: src=%u,%u,%u,%u:%u   dst= %u.%u.%u.%u:%u proto=%d\n",
			       NIPQUAD(exp->tuple.src.ip),ntohs(exp->tuple.src.u.all),
			       NIPQUAD(exp->tuple.dst.ip),ntohs(exp->tuple.dst.u.all),exp->tuple.dst.protonum);
		}
#if 0
		if(ctinfo >= IP_CT_IS_REPLY)
		{/*it is a trick*/
			exp->ct_tuple.dst.protonum = 0;
		}
#endif
		change_ret = ip_conntrack_change_expect(exp, &newtuple);
		DEBUGP_G("change_ret = %d\n", change_ret);
		if(change_ret == 0)
			break;

		if( port >= 65534 )/*just temp solution*/
			port = 30000;
		else 
			port++;
			
			
	}
	
	if (port == 0) 
	{
		return NF_DROP;
	}


	*ret_port = port;

	DEBUGP_G("ip_nat_sdp():  new_ret_port = %d\n",port);

	/*we will add this expect information to table*/
	dataoff = (*pskb)->nh.iph->ihl*4 + sizeof(struct udphdr);

	if (ct_sip_get_info(dptr, (*pskb)->len - dataoff, &matchoff, &matchlen, 
	    &ct_sip_hdrs[POS_CALL_ID_HEADER], &ret_connect_port) > 0)
	{
		struct ip_contrack_sip_rtp *new = NULL;
		DEBUGP_G("find pos_call_id header\n");

		LOCK_BH(&sip_alg_table_lock);
		list_for_each_safe(cur_item, sip_item_temp, &sip_rtp_list) 
		{ 
			sip_item = list_entry(cur_item, struct ip_contrack_sip_rtp, list);
			DEBUGP_G("befor add sip_rtp_list  this sip_item = %p\n", sip_item);
			if(exp == sip_item->this_expect && matchlen == sip_item->callid_length 
				&& strncmp(dptr + matchoff, sip_item->call_id, matchlen) == 0)
			{
				DEBUGP_G("it is a same packet retransmitter Gordon this matched sip_item = %p\n", sip_item);
				UNLOCK_BH(&sip_alg_table_lock);
				goto out;
			}

		}
		UNLOCK_BH(&sip_alg_table_lock);

		new = (struct ip_contrack_sip_rtp *) 
		      kmalloc(sizeof(struct ip_contrack_sip_rtp), GFP_ATOMIC);
		if (!new) {
			DEBUGP_G("sip alg table: OOM allocating expect\n");
			goto out;
		}
		DEBUGP_G("new sip_rtp table = %p\n", new);
		DEBUGP_G("dptr = %p, matchoff = %d, matchlen = %d\n",dptr, matchoff, matchlen);
	
		memset(new, 0, sizeof(struct ip_contrack_sip_rtp));
		INIT_LIST_HEAD(&new->list);
		
		if(matchlen > CALL_ID_LENGTH - 1)
		matchlen = CALL_ID_LENGTH - 1;
		strncpy( new->call_id, dptr + matchoff, matchlen);

		DEBUGP_G("call id is :%s\n",new->call_id);
	
		memcpy(&new->tuple, &exp->tuple, sizeof(exp->tuple));
		memcpy(&new->mask, &exp->mask, sizeof(exp->mask));
		new->this_expect = exp;
		
		new->callid_length = matchlen;

		init_timer(&new->timeout);
		new->timeout.data = (unsigned long)new;
		new->timeout.function = expect_sdp_timed_out;
		new->timeout.expires = jiffies + ip_ct_sip_timeout_invite * HZ;
		add_timer(&new->timeout);

		LOCK_BH(&sip_alg_table_lock);
		list_add(&new->list, &sip_rtp_list);
		UNLOCK_BH(&sip_alg_table_lock);
		
	
	} 
	
out:
	

	if (!mangle_sdp(pskb, ctinfo, ct, newip, port, dptr)) 
	{
		return NF_DROP;
	}
	return NF_ACCEPT;
}

static struct ip_nat_helper sip[MAX_PORTS];
static char sip_names[MAX_PORTS][10];
static void __exit fini(void)
{
	int i;
	struct list_head *cur_item,*sip_item_temp = NULL;
	struct ip_contrack_sip_rtp *sip_item = NULL;
	struct ip_contrack_sip_rtp *sip_del_item = NULL;

	for (i = 0; i < ports_c; i++) {
		DEBUGP("ip_nat_sip: unregistering port %d\n", ports[i]);
		ip_nat_helper_unregister(&sip[i]);
	}
	
	
	
	DEBUGP_G("remove all table, because sip alg module will be removed\n");

		LOCK_BH(&sip_alg_table_lock);

		DEBUGP_G("sip_rtp_list = %p\n",&sip_rtp_list);
		list_for_each_safe(cur_item, sip_item_temp, &sip_rtp_list) 
		{
			DEBUGP_G("cur_item = %p\n",cur_item);
			sip_item = list_entry(cur_item, struct ip_contrack_sip_rtp, list);
			DEBUGP_G("this sip_item = %p, sip_item->list.prev = %p, sip_item->list.next = %p\n", sip_item, sip_item->list.prev, sip_item->list.next);
			
			sip_del_item = sip_item;
			if(sip_del_item)
			{
				if(del_timer(&sip_del_item->timeout))
					DEBUGP_G("Success !!! before delete this sip_item . delete the timer for sip_del_item = %p\n", sip_del_item);
				
			}		
			list_del_init(&sip_del_item->list);
			kfree(sip_del_item);
			sip_del_item = NULL;
		}	

		UNLOCK_BH(&sip_alg_table_lock);

}

static void  fini_err(void)
{
	int i;

	for (i = 0; i < ports_c; i++) {
		DEBUGP("ip_nat_sip: unregistering port %d\n", ports[i]);
		ip_nat_helper_unregister(&sip[i]);
	}
}

static int __init init(void)
{
	int i, ret = 0;
	char *tmpname;

	if (ports[0] == 0)
		ports[0] = SIP_PORT;

	ports[1]=5060;
	for (i = 0; (i < MAX_PORTS) && ports[i]; i++) 
	{

		memset(&sip[i], 0, sizeof(struct ip_nat_helper));

		sip[i].tuple.dst.protonum = IPPROTO_UDP;
		sip[i].tuple.src.u.udp.port = htons(ports[i]);
		sip[i].mask.dst.protonum = 0xFFFF;
		sip[i].mask.src.u.udp.port = 0xFFFF;
		sip[i].help = ip_nat_sip;
		sip[i].me = THIS_MODULE;
		sip[i].flags = IP_NAT_HELPER_F_ALWAYS_IGNORE_EXP;
		sip[i].expect = sip_nat_expected;

		tmpname = &sip_names[i][0];
		if (ports[i] == SIP_PORT)
			sprintf(tmpname, "sip");
		else
			sprintf(tmpname, "sip-%d", i);
		sip[i].name = tmpname;

		DEBUGP("ip_nat_sip: Trying to register for port %d\n",
				ports[i]);
		ret = ip_nat_helper_register(&sip[i]);

		if (ret) 
		{
			DEBUGP("ip_nat_sip: error registering "
			       "helper for port %d\n", ports[i]);
			fini_err();
			return ret;
		}
		ports_c++;
	}

	return ret;
}

module_init(init);
module_exit(fini);
